use crossterm::{
    event::{DisableMouseCapture, EnableMouseCapture, Event, KeyCode, EventStream},
    execute,
    terminal::{disable_raw_mode, enable_raw_mode, EnterAlternateScreen, LeaveAlternateScreen},
};
use ratatui::{prelude::*, widgets::*};
use rumqttc::{AsyncClient, MqttOptions, QoS};
use std::{io, time::Duration, path::PathBuf};
use tokio::sync::mpsc;
use tokio_stream::StreamExt;
use rand::Rng;
use serde::{Deserialize, Serialize};
use configparser::ini::Ini;
use std::fs;

#[derive(Debug, Serialize, Deserialize)]
struct ChatMessage {
    username: String,
    message: String,
}

struct App {
    messages: Vec<(String, String)>,
    input: String,
    username: String,
    current_room: String,
    messages_state: ListState,
}

impl App {
    fn new(username: String) -> App {
        let mut messages = Vec::new();
        let default_room = "default".to_string();
        messages.push(("System".to_string(), format!("Welcome to the chat, {}!", username)));
        messages.push(("System".to_string(), "To change your username, type /nick <new_username>".to_string()));
        messages.push(("System".to_string(), format!("You are currently in room: {}. To change room, type /join <new_room_name>", default_room)));
        let mut messages_state = ListState::default();
        messages_state.select(Some(messages.len().saturating_sub(1)));
        App {
            messages,
            input: String::new(),
            username,
            current_room: default_room,
            messages_state,
        }
    }
}

#[tokio::main]
async fn main() -> Result<(), Box<dyn std::error::Error>> {
    // generate random username
    let mut rng = rand::thread_rng();
    let random_number = rng.gen_range(1..=999);
    let username = format!("User{}", random_number);

    // setup terminal
    enable_raw_mode()?;
    let mut stdout = io::stdout();
    execute!(stdout, EnterAlternateScreen, EnableMouseCapture)?;
    let backend = CrosstermBackend::new(stdout);
    let mut terminal = Terminal::new(backend)?;

    // create app and run it
    let app = App::new(username);
    let res = run_app(&mut terminal, app).await;

    // restore terminal
    disable_raw_mode()?;
    execute!(
        terminal.backend_mut(),
        LeaveAlternateScreen,
        DisableMouseCapture
    )?;
    terminal.show_cursor()?;

    if let Err(err) = res {
        println!("{err:?}");
    }

    Ok(())
}

async fn run_app<B: Backend>(terminal: &mut Terminal<B>, mut app: App) -> io::Result<()> {
    let (tx, mut rx) = mpsc::channel::<(String, String)>(100);

    let mut config = Ini::new();
    let mut config_path = PathBuf::new();
    if let Some(home_dir) = dirs::home_dir() {
        config_path.push(home_dir);
        config_path.push(".config");
        fs::create_dir_all(&config_path).unwrap();
        config_path.push("mqtt-chat.ini");
    }

    if !config_path.exists() {
        config.set("mqtt", "server", Some("172.16.0.3".to_owned()));
        config.set("mqtt", "port", Some("1883".to_owned()));
        config.set("mqtt", "username", Some("".to_owned()));
        config.set("mqtt", "password", Some("".to_owned()));
        config.write(&config_path).unwrap();
    }

    let map = config.load(&config_path).unwrap();
    let server = map.get("mqtt").unwrap().get("server").unwrap().clone().unwrap();
    let port = map.get("mqtt").unwrap().get("port").unwrap().clone().unwrap().parse::<u16>().unwrap();
    let username = map.get("mqtt").unwrap().get("username").unwrap().clone().unwrap();
    let password = map.get("mqtt").unwrap().get("password").unwrap().clone().unwrap();

    let mut mqttoptions = MqttOptions::new(app.username.clone(), server, port);
    mqttoptions.set_keep_alive(Duration::from_secs(5));
    if !username.is_empty() {
        mqttoptions.set_credentials(username, password);
    }

    let (client, mut eventloop) = AsyncClient::new(mqttoptions, 10);
    client.subscribe(&format!("chat/{}", app.current_room), QoS::AtMostOnce).await.unwrap();

    // Publish connect message
    let connect_message = ChatMessage {
        username: app.username.clone(),
        message: "has connected".to_string(),
    };
    let connect_payload = serde_json::to_string(&connect_message).unwrap();
    client.publish(&format!("chat/{}", app.current_room), QoS::AtMostOnce, false, connect_payload.as_bytes()).await.unwrap();

    let client_clone = client.clone();

    tokio::spawn(async move {
        while let Ok(notification) = eventloop.poll().await {
            if let rumqttc::Event::Incoming(rumqttc::Packet::Publish(p)) = notification {
                let message_str = String::from_utf8_lossy(&p.payload).to_string();
                match serde_json::from_str::<ChatMessage>(&message_str) {
                    Ok(chat_msg) => {
                        if tx.send((chat_msg.username, chat_msg.message)).await.is_err() {
                            break;
                        }
                    }
                    Err(_) => {
                        // Handle malformed messages
                        if tx.send(("System".to_string(), message_str)).await.is_err() {
                            break;
                        }
                    }
                }
            }
        }
    });

    let mut reader = EventStream::new();

    loop {
        terminal.draw(|f| ui(f, &mut app))?;

        tokio::select! {
            Some(Ok(event)) = reader.next() => {
                if let Event::Key(key) = event {
                    match key.code {
                        KeyCode::Enter => {
                            let message_text = app.input.drain(..).collect::<String>();
                            if message_text.starts_with("/nick ") {
                                let old_username = app.username.clone();
                                let new_username = message_text.split_whitespace().nth(1).unwrap_or(&app.username).to_string();
                                app.username = new_username.clone();
                                app.messages.push(("System".to_string(), format!("Username changed to: {}", app.username)));
                                app.messages_state.select(Some(app.messages.len().saturating_sub(1)));

                                let chat_message = ChatMessage {
                                    username: old_username,
                                    message: format!("has changed their name to {}", new_username),
                                };
                                let payload = serde_json::to_string(&chat_message).unwrap();
                                client_clone
                                    .publish(&format!("chat/{}", app.current_room), QoS::AtMostOnce, false, payload.as_bytes())
                                    .await
                                    .unwrap();
                            } else if message_text.starts_with("/join ") {
                                let old_room = app.current_room.clone();
                                let new_room = message_text.split_whitespace().nth(1).unwrap_or(&app.current_room).to_string();
                                if old_room != new_room {
                                    // Send leaving message to old room
                                    let leaving_message = ChatMessage {
                                        username: app.username.clone(),
                                        message: format!("is leaving for room {}", new_room),
                                    };
                                    let leaving_payload = serde_json::to_string(&leaving_message).unwrap();
                                    client_clone.publish(&format!("chat/{}", old_room), QoS::AtMostOnce, false, leaving_payload.as_bytes()).await.unwrap();

                                    client_clone.unsubscribe(&format!("chat/{}", old_room)).await.unwrap();
                                    client_clone.subscribe(&format!("chat/{}", new_room), QoS::AtMostOnce).await.unwrap();
                                    app.current_room = new_room.clone();
                                    app.messages.push(("System".to_string(), format!("Changed room to: {}", new_room)));
                                    app.messages_state.select(Some(app.messages.len().saturating_sub(1)));

                                    // Send joining message to new room
                                    let joining_message = ChatMessage {
                                        username: app.username.clone(),
                                        message: "has joined the room".to_string(),
                                    };
                                    let joining_payload = serde_json::to_string(&joining_message).unwrap();
                                    client_clone.publish(&format!("chat/{}", new_room), QoS::AtMostOnce, false, joining_payload.as_bytes()).await.unwrap();
                                } else {
                                    app.messages.push(("System".to_string(), format!("Already in room: {}", new_room)));
                                    app.messages_state.select(Some(app.messages.len().saturating_sub(1)));
                                }
                            } else {
                                let chat_message = ChatMessage {
                                    username: app.username.clone(),
                                    message: message_text,
                                };
                                let payload = serde_json::to_string(&chat_message).unwrap();
                                client_clone
                                    .publish(&format!("chat/{}", app.current_room), QoS::AtMostOnce, false, payload.as_bytes())
                                    .await
                                    .unwrap();
                            }
                        }
                        KeyCode::Char(c) => {
                            app.input.push(c);
                        }
                        KeyCode::Backspace => {
                            app.input.pop();
                        }
                        KeyCode::Esc => {
                            return Ok(());
                        }
                        _ => {}
                    }
                }
            }
            Some((username, message)) = rx.recv() => {
                app.messages.push((username, message));
                app.messages_state.select(Some(app.messages.len().saturating_sub(1)));
            }
            else => {
                break;
            }
        }
    }

    Ok(())
}

fn ui(f: &mut Frame, app: &mut App) {
    let chunks = Layout::default()
        .direction(Direction::Vertical)
        .margin(1)
        .constraints([Constraint::Percentage(80), Constraint::Percentage(20)].as_ref())
        .split(f.size());

    let messages: Vec<ListItem> = app
        .messages
        .iter()
        .map(|(username, message)| ListItem::new(format!("{}: {}", username, message)))
        .collect();
    let messages_list = List::new(messages)
        .block(Block::default().borders(Borders::ALL).title("Messages"));
    f.render_stateful_widget(messages_list, chunks[0], &mut app.messages_state);

    let input = Paragraph::new(app.input.as_str())
        .style(Style::default())
        .block(Block::default().borders(Borders::ALL).title("Input"));
    f.render_widget(input, chunks[1]);
}
